from imgcv.common import check_image
import numpy as np


def connected_component_analysis(im):
    """Peforms connected component analysis on a binary image. Currently only 4-connected component analysis is supported.

    NOTE: The input image should be a binary image. Non-zero pixels are considered as foreground. This function will not work properly if the input image is not binary.

    Args:
        im (np.ndarray): Binary image. Non-zero pixels are considered as foreground

    Raises:
        ValueError: If the input image is not binary

    Returns:
        np.ndarray: Image with connected components labeled. Each connected component will have a unique label
        int: Number of connected components in the image
    """

    # check if the input image is valid
    check_image(im)

    # check if the image is binary
    if not _is_binary(im):
        raise ValueError("The input image is not binary")

    connected_components = np.zeros_like(im, dtype=np.uint8)

    window = np.zeros((2, 2), dtype=np.uint8)

    # padding the image
    padded_im = np.pad(im, 1, mode="constant")

    label_counter = 1
    equivalence = {}

    # first pass
    for row in range(padded_im.shape[0] - 1):
        for col in range(padded_im.shape[1] - 1):
            window = padded_im[row : row + 2, col : col + 2]
            main_pixel = window[
                1, 1
            ]  # bottom right corner will be our main pixel as our image is padded

            # We are only interested in those pixels whose intensity is 1. So, if the main pixel is 0, then we will skip
            if main_pixel == 0:
                continue

            # if both left and top pixels are 0, then assign a new label
            if window[0, 1] == 0 and window[1, 0] == 0:
                connected_components[row, col] = label_counter
                equivalence[label_counter] = label_counter
                label_counter += 1

            # if both are not 0, then assign the minimum label and update the equivalence dictionary
            elif window[1, 0] != 0 and window[0, 1] != 0:
                connected_components[row, col] = min(
                    connected_components[row - 1, col],
                    connected_components[row, col - 1],
                )
                equivalence[
                    max(
                        connected_components[row - 1, col],
                        connected_components[row, col - 1],
                    )
                ] = min(
                    connected_components[row - 1, col],
                    connected_components[row, col - 1],
                )

            # if any one of the pixel is 0, then assign the non-zero label
            else:
                connected_components[row, col] = max(
                    connected_components[row - 1, col],
                    connected_components[row, col - 1],
                )
                equivalence[
                    max(
                        connected_components[row - 1, col],
                        connected_components[row, col - 1],
                    )
                ] = max(
                    connected_components[row - 1, col],
                    connected_components[row, col - 1],
                )

    # second pass
    # replace the labels with the equivalent labels
    for key, value in equivalence.items():
        connected_components[connected_components == key] = value

    # Number of connected components
    no_of_connected_components = len(np.unique(connected_components)) - 1

    """
    connected_components includes the labels of the connected components. The user can now use this in any way they want. For example, they can find the number of connected components, the area of each connected component, etc. They can even extract the connected components from the original image using these labels.
    """
    return connected_components, no_of_connected_components


def _is_binary(im):
    """Checks if the image is binary or not. A binary image is an image that has only two pixel values. For example, 0 and 1.

    Returns:
        Boolean: True if the image is binary, False otherwise
    """

    pixels = np.unique(im)
    return np.array_equal(pixels, np.array([0, 1]))



import numpy as np
from imgcv.common import check_image


def reduce_gray_levels(image, k):
    """Reduce the number of gray levels in the input image to desired levels

    Args:
        image (np.ndarray): Input image to be transformed
        k (int): Desired number of bits to represent the image. The number of intensity levels will be 2^k. For example, if k=2, the number of intensity levels will be 4.

    Raises:
        TypeError: If the input is not a numpy array
        ValueError: If the input is not a grayscale or color image

    Returns:
        np.ndarray: Image with reduced gray levels
    """

    check_image(image)

    if len(image.shape) == 2:
        intensity_levels = 2**k
        compression_ratio = 256 / intensity_levels
        reduced_image = np.floor(image / compression_ratio) * compression_ratio
        return reduced_image
    else:
        # need to tackle each channel separately
        reduced_image = np.zeros_like(image)
        for i in range(image.shape[2]):
            reduced_image[:, :, i] = reduce_gray_levels(image[:, :, i], k)

        return reduced_image



import numpy as np
from imgcv.common import check_image


def gamma_correction(image, gamma=1.0, c=1):
    """Applies gamma correction to the image.

    Formula:
        gamma_correction = c * (image ** gamma)

    Args:
        image (np.ndarray): Image to be transformed.
        gamma (float, optional): Power value . Defaults to 1.0.
        c (int, optional): Constant value. Defaults to 1.

    Returns:
        np.ndarray: Transformed image.
    """

    check_image(image)

    norm_img = image / 255.0
    corrected_img = c * (norm_img**gamma)

    # Clip the values to be in the range [0, 1] since gamma correction can result in overflow
    corrected_img = np.clip(corrected_img, 0, 1)
    # scale the values back to [0, 255]
    corrected_img = (corrected_img * 255).astype(np.uint8)

    return corrected_img



from imgcv.common import check_image


def to_binary(im, threshold=127):
    """Converts an image to binary image

    Args:
        im (np.ndarray): Input image
        threshold (int): Threshold value

    Returns:
        np.ndarray: Binary image
    """

    check_image(im)

    im[im > threshold] = 255
    im[im <= threshold] = 0

    # convert 255 to 1
    im[im == 255] = 1

    return im
